@@ -37,7 +37,7 @@ if STDERR_FILE:
   logging.info('logging stderr in js compiler phase into %s' % STDERR_FILE)
   STDERR_FILE = open(STDERR_FILE, 'w')
 
-def emscript(infile, settings, outfile, libraries=[], compiler_engine=None,
+def emscript(infile, settings, outfile, outfile_name, libraries=[], compiler_engine=None,
              temp_files=None, DEBUG=None, DEBUG_CACHE=None):
   """Runs the emscripten LLVM-to-JS compiler.
 
@@ -177,6 +177,26 @@ def emscript(infile, settings, outfile, libraries=[], compiler_engine=None,
 
     metadata['declares'] = filter(lambda i64_func: i64_func not in ['getHigh32', 'setHigh32', '__muldi3', '__divdi3', '__remdi3', '__udivdi3', '__uremdi3'], metadata['declares']) # FIXME: do these one by one as normal js lib funcs
 
+    # Syscalls optimization. Our syscalls are static, and so if we see a very limited set of them - in particular,
+    # no open() syscall and just simple writing - then we don't need full filesystem support.
+    # If FORCE_FILESYSTEM is set, we can't do this. We also don't do it if INCLUDE_FULL_LIBRARY, since
+    # not including the filesystem would mean not including the full JS libraries, and the same for
+    # MAIN_MODULE since a side module might need the filesystem.
+    if not settings['NO_FILESYSTEM'] and not settings['FORCE_FILESYSTEM'] and not settings['INCLUDE_FULL_LIBRARY'] and not settings['MAIN_MODULE']:
+      syscall_prefix = '__syscall'
+      syscalls = filter(lambda declare: declare.startswith(syscall_prefix), metadata['declares'])
+      def is_int(x):
+        try:
+          int(x)
+          return True
+        except:
+          return False
+      syscalls = filter(lambda declare: is_int(declare[len(syscall_prefix):]), syscalls)
+      syscalls = map(lambda declare: int(declare[len(syscall_prefix):]), syscalls)
+      if set(syscalls).issubset(set([6, 54, 140, 146])): # close, ioctl, llseek, writev
+        if DEBUG: logging.debug('very limited syscalls (%s) so disabling full filesystem support' % ', '.join(map(str, syscalls)))
+        settings['NO_FILESYSTEM'] = 1
+
     # Integrate info from backend
     if settings['SIDE_MODULE']:
       settings['DEFAULT_LIBRARY_FUNCS_TO_INCLUDE'] = [] # we don't need any JS library contents in side modules
@@ -193,6 +213,8 @@ def emscript(infile, settings, outfile, libraries=[], compiler_engine=None,
 
     settings['MAX_GLOBAL_ALIGN'] = metadata['maxGlobalAlign']
 
+    settings['IMPLEMENTED_FUNCTIONS'] = metadata['implementedFunctions']
+
     assert not (metadata['simd'] and settings['SPLIT_MEMORY']), 'SIMD is used, but not supported in SPLIT_MEMORY'
 
     # Save settings to a file to work around v8 issue 1579
@@ -245,7 +267,9 @@ def emscript(infile, settings, outfile, libraries=[], compiler_engine=None,
            mem_init))
 
     if settings['SIDE_MODULE']:
-      pre = pre.replace('Runtime.GLOBAL_BASE', 'gb').replace('{{{ STATIC_BUMP }}}', str(staticbump))
+      pre = pre.replace('Runtime.GLOBAL_BASE', 'gb')
+    if settings['SIDE_MODULE'] or settings['BINARYEN']:
+      pre = pre.replace('{{{ STATIC_BUMP }}}', str(staticbump))
 
     funcs_js = [funcs]
     parts = pre.split('// ASM_LIBRARY FUNCTIONS\n')
@@ -473,35 +497,52 @@ function _emscripten_asm_const_%s(%s) {
     maths = ['Math.' + func for func in ['floor', 'abs', 'sqrt', 'pow', 'cos', 'sin', 'tan', 'acos', 'asin', 'atan', 'atan2', 'exp', 'log', 'ceil', 'imul', 'min', 'clz32']]
     simdfloattypes = []
     simdinttypes = []
-    simdfuncs = ['check', 'add', 'sub', 'neg', 'mul',
-                 'equal', 'lessThan', 'greaterThan',
-                 'notEqual', 'lessThanOrEqual', 'greaterThanOrEqual',
-                 'select', 'and', 'or', 'xor', 'not',
-                 'splat', 'swizzle', 'shuffle',
-                 'load', 'store', 'load1', 'store1', 'load2', 'store2', 'load3', 'store3',
-                 'extractLane', 'replaceLane']
+    simdbooltypes = []
+    simdfuncs = ['splat', 'check', 'extractLane', 'replaceLane']
+    simdintfloatfuncs = ['add', 'sub', 'neg', 'mul',
+                         'equal', 'lessThan', 'greaterThan',
+                         'notEqual', 'lessThanOrEqual', 'greaterThanOrEqual',
+                         'select', 'swizzle', 'shuffle',
+                         'load', 'store', 'load1', 'store1', 'load2', 'store2', 'load3', 'store3']
+    simdintboolfuncs = ['and', 'xor', 'or', 'not']
+    if metadata['simdUint8x16']:
+      simdinttypes += ['Uint8x16']
+      simdintfloatfuncs += ['fromUint8x16Bits']
     if metadata['simdInt8x16']:
       simdinttypes += ['Int8x16']
-      simdfuncs += ['fromInt8x16Bits']
+      simdintfloatfuncs += ['fromInt8x16Bits']
+    if metadata['simdUint16x8']:
+      simdinttypes += ['Uint16x8']
+      simdintfloatfuncs += ['fromUint16x8Bits']
     if metadata['simdInt16x8']:
       simdinttypes += ['Int16x8']
-      simdfuncs += ['fromInt16x8Bits']
+      simdintfloatfuncs += ['fromInt16x8Bits']
+    if metadata['simdUint32x4']:
+      simdinttypes += ['Uint32x4']
+      simdintfloatfuncs += ['fromUint32x4', 'fromUint32x4Bits']
     if metadata['simdInt32x4']:
       simdinttypes += ['Int32x4']
-      simdfuncs += ['fromInt32x4', 'fromInt32x4Bits']
+      simdintfloatfuncs += ['fromInt32x4', 'fromInt32x4Bits']
     if metadata['simdFloat32x4']:
       simdfloattypes += ['Float32x4']
-      simdfuncs += ['fromFloat32x4', 'fromFloat32x4Bits']
+      simdintfloatfuncs += ['fromFloat32x4', 'fromFloat32x4Bits']
     if metadata['simdFloat64x2']:
       simdfloattypes += ['Float64x2']
-      simdfuncs += ['fromFloat64x2', 'fromFloat64x2Bits']
+      simdintfloatfuncs += ['fromFloat64x2', 'fromFloat64x2Bits']
+    if metadata['simdBool8x16']:
+      simdbooltypes += ['Bool8x16']
+    if metadata['simdBool16x8']:
+      simdbooltypes += ['Bool16x8']
+    if metadata['simdBool32x4']:
+      simdbooltypes += ['Bool32x4']
+    if metadata['simdBool64x2']:
+      simdbooltypes += ['Bool64x2']
 
-    simdfloatfuncs = simdfuncs + ['div', 'min', 'max', 'minNum', 'maxNum', 'sqrt',
-                                  'abs', 'reciprocalApproximation', 'reciprocalSqrtApproximation'];
-    simdintfuncs = simdfuncs + ['shiftRightArithmeticByScalar',
-                                'shiftRightLogicalByScalar',
-                                'shiftLeftByScalar'];
-    simdtypes = simdfloattypes + simdinttypes
+    simdfloatfuncs = simdfuncs + simdintfloatfuncs + ['div', 'min', 'max', 'minNum', 'maxNum', 'sqrt',
+                                  'abs', 'reciprocalApproximation', 'reciprocalSqrtApproximation']
+    simdintfuncs = simdfuncs + simdintfloatfuncs + simdintboolfuncs + ['shiftLeftByScalar', 'shiftRightByScalar', 'addSaturate', 'subSaturate']
+    simdboolfuncs = simdfuncs + simdintboolfuncs + ['anyTrue', 'allTrue']
+    simdtypes = simdfloattypes + simdinttypes + simdbooltypes
 
     fundamentals = ['Math']
     fundamentals += ['Int8Array', 'Int16Array', 'Int32Array', 'Uint8Array', 'Uint16Array', 'Uint32Array', 'Float32Array', 'Float64Array']
@@ -547,6 +588,7 @@ function _emscripten_asm_const_%s(%s) {
              '"); ' + extra
 
     basic_funcs = ['abort', 'assert'] + [m.replace('.', '_') for m in math_envs]
+    if settings['STACK_OVERFLOW_CHECK']: basic_funcs += ['abortStackOverflow']
 
     asm_safe_heap = settings['SAFE_HEAP'] and not settings['SAFE_HEAP_LOG'] and not settings['RELOCATABLE'] # optimized safe heap in asm, when we can
 
@@ -599,6 +641,9 @@ function _emscripten_asm_const_%s(%s) {
     if settings['SAFE_HEAP']:
       asm_runtime_funcs += ['setDynamicTop']
 
+    if settings['ONLY_MY_CODE']:
+      asm_runtime_funcs = []
+
     # function tables
     if not settings['EMULATED_FUNCTION_POINTERS']:
       function_tables = ['dynCall_' + table for table in last_forwarded_json['Functions']['tables']]
@@ -670,19 +715,20 @@ function ftCall_%s(%s) {%s
 
     def quote(prop):
       if settings['USE_CLOSURE_COMPILER'] == 2:
-        return "'" + prop + "'"
+        return ''.join(map(lambda p: "'" + p + "'", prop.split('.')))
       else:
         return prop
 
     def access_quote(prop):
       if settings['USE_CLOSURE_COMPILER'] == 2:
-        return "['" + prop + "']"
+        return ''.join(map(lambda p: "['" + p + "']", prop.split('.')))
       else:
         return '.' + prop
 
     # calculate exports
     exported_implemented_functions = list(exported_implemented_functions) + metadata['initializers']
-    exported_implemented_functions.append('runPostSets')
+    if not settings['ONLY_MY_CODE']:
+      exported_implemented_functions.append('runPostSets')
     if settings['ALLOW_MEMORY_GROWTH']:
       exported_implemented_functions.append('_emscripten_replace_memory')
     all_exported = exported_implemented_functions + asm_runtime_funcs + function_tables
@@ -721,7 +767,10 @@ function ftCall_%s(%s) {%s
           if sub in s:
             return True
         return False
-      nonexisting_simd_symbols = ['Int32x4_fromInt32x4', 'Float32x4_fromFloat32x4']
+      nonexisting_simd_symbols = ['Int8x16_fromInt8x16', 'Uint8x16_fromUint8x16', 'Int16x8_fromInt16x8', 'Uint16x8_fromUint16x8', 'Int32x4_fromInt32x4', 'Uint32x4_fromUint32x4', 'Float32x4_fromFloat32x4', 'Float64x2_fromFloat64x2']
+      nonexisting_simd_symbols += ['Int32x4_addSaturate', 'Int32x4_subSaturate', 'Uint32x4_addSaturate', 'Uint32x4_subSaturate']
+      nonexisting_simd_symbols += [(x + '_' + y) for x in ['Int8x16', 'Uint8x16', 'Int16x8', 'Uint16x8', 'Float64x2'] for y in ['load2', 'load3', 'store2', 'store3']]
+      nonexisting_simd_symbols += [(x + '_' + y) for x in ['Int8x16', 'Uint8x16', 'Int16x8', 'Uint16x8'] for y in ['load1', 'store1']]
 
       asm_global_funcs += ''.join(['  var SIMD_' + ty + '=global' + access_quote('SIMD') + access_quote(ty) + ';\n' for ty in simdtypes])
 
@@ -732,13 +781,16 @@ function ftCall_%s(%s) {%s
       simd_float_symbols = ['  var SIMD_' + ty + '_' + g + '=SIMD_' + ty + access_quote(g) + ';\n' for ty in simdfloattypes for g in simdfloatfuncs]
       simd_float_symbols = filter(lambda x: not string_contains_any(x, nonexisting_simd_symbols), simd_float_symbols)
       asm_global_funcs += ''.join(simd_float_symbols)
+
+      simd_bool_symbols = ['  var SIMD_' + ty + '_' + g + '=SIMD_' + ty + access_quote(g) + ';\n' for ty in simdbooltypes for g in simdboolfuncs]
+      simd_bool_symbols = filter(lambda x: not string_contains_any(x, nonexisting_simd_symbols), simd_bool_symbols)
+      asm_global_funcs += ''.join(simd_bool_symbols)
+
       # Unofficial, Bool64x2 does not yet exist, but needed for Float64x2 comparisons.
       if metadata['simdFloat64x2']:
         asm_global_funcs += '  var SIMD_Int32x4_fromBool64x2Bits = global.SIMD.Int32x4.fromBool64x2Bits;\n';
     if settings['USE_PTHREADS']:
-#      asm_global_funcs += ''.join(['  var Atomics_' + ty + '=global' + access_quote('Atomics') + access_quote(ty) + ';\n' for ty in ['load', 'store', 'exchange', 'compareExchange', 'add', 'sub', 'and', 'or', 'xor', 'fence']])
-# TODO: Once bug https://bugzilla.mozilla.org/show_bug.cgi?id=1141986 is implemented, replace the following line with the above one!
-      asm_global_funcs += ''.join(['  var Atomics_' + ty + '=global' + access_quote('Atomics') + access_quote(ty) + ';\n' for ty in ['load', 'store', 'compareExchange', 'add', 'sub', 'and', 'or', 'xor', 'fence']])
+      asm_global_funcs += ''.join(['  var Atomics_' + ty + '=global' + access_quote('Atomics') + access_quote(ty) + ';\n' for ty in ['load', 'store', 'exchange', 'compareExchange', 'add', 'sub', 'and', 'or', 'xor']])
     asm_global_vars = ''.join(['  var ' + g + '=env' + access_quote(g) + '|0;\n' for g in basic_vars + global_vars])
 
     # sent data
@@ -795,16 +847,7 @@ Module['NAMED_GLOBALS'] = NAMED_GLOBALS;
       receiving += ''.join(["Module['%s'] = Module['%s']\n" % (k, v) for k, v in metadata['aliases'].iteritems()])
 
     if settings['USE_PTHREADS']:
-      shared_array_buffer = '''if (typeof SharedArrayBuffer !== 'undefined') Module.asmGlobalArg['Atomics'] = Atomics;
-if (typeof SharedInt8Array !== 'undefined') Module.asmGlobalArg['SharedInt8Array'] = SharedInt8Array;
-if (typeof SharedInt16Array !== 'undefined') Module.asmGlobalArg['SharedInt16Array'] = SharedInt16Array;
-if (typeof SharedInt32Array !== 'undefined') Module.asmGlobalArg['SharedInt32Array'] = SharedInt32Array;
-if (typeof SharedUint8Array !== 'undefined') Module.asmGlobalArg['SharedUint8Array'] = SharedUint8Array;
-if (typeof SharedUint16Array !== 'undefined') Module.asmGlobalArg['SharedUint16Array'] = SharedUint16Array;
-if (typeof SharedUint32Array !== 'undefined') Module.asmGlobalArg['SharedUint32Array'] = SharedUint32Array;
-if (typeof SharedFloat32Array !== 'undefined') Module.asmGlobalArg['SharedFloat32Array'] = SharedFloat32Array;
-if (typeof SharedFloat64Array !== 'undefined') Module.asmGlobalArg['SharedFloat64Array'] = SharedFloat64Array;
-'''
+      shared_array_buffer = "if (typeof SharedArrayBuffer !== 'undefined') Module.asmGlobalArg['Atomics'] = Atomics;"
     else:
       shared_array_buffer = ''
 
@@ -887,99 +930,16 @@ function setF64(ptr, value) {
 '''
       first_in_asm += 'buffer = new ArrayBuffer(32); // fake\n'
 
-    funcs_js = ['''
-%s
-Module%s = %s;
-%s
-Module%s = %s;
-// EMSCRIPTEN_START_ASM
-var asm = (function(global, env, buffer) {
-  %s
-  %s
-  %s
-''' % (asm_setup,
-       access_quote('asmGlobalArg'), the_global,
-       shared_array_buffer,
-       access_quote('asmLibraryArg'), sending,
-       "'use asm';" if not metadata.get('hasInlineJS') and settings['ASM_JS'] == 1 else "'almost asm';", 
-       first_in_asm,
-       '''
-  var HEAP8 = new global%s(buffer);
-  var HEAP16 = new global%s(buffer);
-  var HEAP32 = new global%s(buffer);
-  var HEAPU8 = new global%s(buffer);
-  var HEAPU16 = new global%s(buffer);
-  var HEAPU32 = new global%s(buffer);
-  var HEAPF32 = new global%s(buffer);
-  var HEAPF64 = new global%s(buffer);
-''' % (access_quote('SharedInt8Array' if settings['USE_PTHREADS'] else 'Int8Array'),
-     access_quote('SharedInt16Array' if settings['USE_PTHREADS'] else 'Int16Array'),
-     access_quote('SharedInt32Array' if settings['USE_PTHREADS'] else 'Int32Array'),
-     access_quote('SharedUint8Array' if settings['USE_PTHREADS'] else 'Uint8Array'),
-     access_quote('SharedUint16Array' if settings['USE_PTHREADS'] else 'Uint16Array'),
-     access_quote('SharedUint32Array' if settings['USE_PTHREADS'] else 'Uint32Array'),
-     access_quote('SharedFloat32Array' if settings['USE_PTHREADS'] else 'Float32Array'),
-     access_quote('SharedFloat64Array' if settings['USE_PTHREADS'] else 'Float64Array'))
-     if not settings['ALLOW_MEMORY_GROWTH'] else '''
-  var Int8View = global%s;
-  var Int16View = global%s;
-  var Int32View = global%s;
-  var Uint8View = global%s;
-  var Uint16View = global%s;
-  var Uint32View = global%s;
-  var Float32View = global%s;
-  var Float64View = global%s;
-  var HEAP8 = new Int8View(buffer);
-  var HEAP16 = new Int16View(buffer);
-  var HEAP32 = new Int32View(buffer);
-  var HEAPU8 = new Uint8View(buffer);
-  var HEAPU16 = new Uint16View(buffer);
-  var HEAPU32 = new Uint32View(buffer);
-  var HEAPF32 = new Float32View(buffer);
-  var HEAPF64 = new Float64View(buffer);
-  var byteLength = global.byteLength;
-''' % (access_quote('Int8Array'),
-     access_quote('Int16Array'),
-     access_quote('Int32Array'),
-     access_quote('Uint8Array'),
-     access_quote('Uint16Array'),
-     access_quote('Uint32Array'),
-     access_quote('Float32Array'),
-     access_quote('Float64Array'))) + '\n' + asm_global_vars + ('''
-  var __THREW__ = 0;
-  var threwValue = 0;
-  var setjmpId = 0;
-  var undef = 0;
-  var nan = global%s, inf = global%s;
-  var tempInt = 0, tempBigInt = 0, tempBigIntP = 0, tempBigIntS = 0, tempBigIntR = 0.0, tempBigIntI = 0, tempBigIntD = 0, tempValue = 0, tempDouble = 0.0;
-''' % (access_quote('NaN'), access_quote('Infinity'))) + ''.join(['''
-  var tempRet%d = 0;''' % i for i in range(10)]) + '\n' + asm_global_funcs] + \
-  ['  var tempFloat = %s;\n' % ('Math_fround(0)' if provide_fround else '0.0')] + \
-  ['  var asyncState = 0;\n' if settings.get('EMTERPRETIFY_ASYNC') else ''] + \
-  (['  const f0 = Math_fround(0);\n'] if provide_fround else []) + \
-  ['' if not settings['ALLOW_MEMORY_GROWTH'] else '''
-function _emscripten_replace_memory(newBuffer) {
-  if ((byteLength(newBuffer) & 0xffffff || byteLength(newBuffer) <= 0xffffff) || byteLength(newBuffer) > 0x80000000) return false;
-  HEAP8 = new Int8View(newBuffer);
-  HEAP16 = new Int16View(newBuffer);
-  HEAP32 = new Int32View(newBuffer);
-  HEAPU8 = new Uint8View(newBuffer);
-  HEAPU16 = new Uint16View(newBuffer);
-  HEAPU32 = new Uint32View(newBuffer);
-  HEAPF32 = new Float32View(newBuffer);
-  HEAPF64 = new Float64View(newBuffer);
-  buffer = newBuffer;
-  return true;
-}
-'''] + ['''
-// EMSCRIPTEN_START_FUNCS
+    runtime_funcs = []
+    if not settings['ONLY_MY_CODE']:
+      runtime_funcs = ['''
 function stackAlloc(size) {
   size = size|0;
   var ret = 0;
   ret = STACKTOP;
   STACKTOP = (STACKTOP + size)|0;
   STACKTOP = (STACKTOP + 15)&-16;
-''' + ('if ((STACKTOP|0) >= (STACK_MAX|0)) abort();\n' if settings['ASSERTIONS'] else '') + '''
+''' + ('  if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(size|0);\n' if (settings['ASSERTIONS'] or settings['STACK_OVERFLOW_CHECK'] >= 2) else '') + '''
   return ret|0;
 }
 function stackSave() {
@@ -1126,7 +1086,95 @@ function setTempRet0(value) {
 function getTempRet0() {
   return tempRet0|0;
 }
-''' if not settings['RELOCATABLE'] else ''] + funcs_js + ['''
+''' if not settings['RELOCATABLE'] else '']
+
+    funcs_js = ['''
+%s
+Module%s = %s;
+%s
+Module%s = %s;
+// EMSCRIPTEN_START_ASM
+var asm = (function(global, env, buffer) {
+  %s
+  %s
+  %s
+''' % (asm_setup,
+       access_quote('asmGlobalArg'), the_global,
+       shared_array_buffer,
+       access_quote('asmLibraryArg'), sending,
+       "'use asm';" if not metadata.get('hasInlineJS') and settings['ASM_JS'] == 1 else "'almost asm';", 
+       first_in_asm,
+       '''
+  var HEAP8 = new global%s(buffer);
+  var HEAP16 = new global%s(buffer);
+  var HEAP32 = new global%s(buffer);
+  var HEAPU8 = new global%s(buffer);
+  var HEAPU16 = new global%s(buffer);
+  var HEAPU32 = new global%s(buffer);
+  var HEAPF32 = new global%s(buffer);
+  var HEAPF64 = new global%s(buffer);
+''' % (access_quote('Int8Array'),
+     access_quote('Int16Array'),
+     access_quote('Int32Array'),
+     access_quote('Uint8Array'),
+     access_quote('Uint16Array'),
+     access_quote('Uint32Array'),
+     access_quote('Float32Array'),
+     access_quote('Float64Array'))
+     if not settings['ALLOW_MEMORY_GROWTH'] else '''
+  var Int8View = global%s;
+  var Int16View = global%s;
+  var Int32View = global%s;
+  var Uint8View = global%s;
+  var Uint16View = global%s;
+  var Uint32View = global%s;
+  var Float32View = global%s;
+  var Float64View = global%s;
+  var HEAP8 = new Int8View(buffer);
+  var HEAP16 = new Int16View(buffer);
+  var HEAP32 = new Int32View(buffer);
+  var HEAPU8 = new Uint8View(buffer);
+  var HEAPU16 = new Uint16View(buffer);
+  var HEAPU32 = new Uint32View(buffer);
+  var HEAPF32 = new Float32View(buffer);
+  var HEAPF64 = new Float64View(buffer);
+  var byteLength = global.byteLength;
+''' % (access_quote('Int8Array'),
+     access_quote('Int16Array'),
+     access_quote('Int32Array'),
+     access_quote('Uint8Array'),
+     access_quote('Uint16Array'),
+     access_quote('Uint32Array'),
+     access_quote('Float32Array'),
+     access_quote('Float64Array'))) + '\n' + asm_global_vars + ('''
+  var __THREW__ = 0;
+  var threwValue = 0;
+  var setjmpId = 0;
+  var undef = 0;
+  var nan = global%s, inf = global%s;
+  var tempInt = 0, tempBigInt = 0, tempBigIntP = 0, tempBigIntS = 0, tempBigIntR = 0.0, tempBigIntI = 0, tempBigIntD = 0, tempValue = 0, tempDouble = 0.0;
+''' % (access_quote('NaN'), access_quote('Infinity'))) + ''.join(['''
+  var tempRet%d = 0;''' % i for i in range(10)]) + '\n' + asm_global_funcs] + \
+  ['  var tempFloat = %s;\n' % ('Math_fround(0)' if provide_fround else '0.0')] + \
+  ['  var asyncState = 0;\n' if settings.get('EMTERPRETIFY_ASYNC') else ''] + \
+  (['  const f0 = Math_fround(0);\n'] if provide_fround else []) + \
+  ['' if not settings['ALLOW_MEMORY_GROWTH'] else '''
+function _emscripten_replace_memory(newBuffer) {
+  if ((byteLength(newBuffer) & 0xffffff || byteLength(newBuffer) <= 0xffffff) || byteLength(newBuffer) > 0x80000000) return false;
+  HEAP8 = new Int8View(newBuffer);
+  HEAP16 = new Int16View(newBuffer);
+  HEAP32 = new Int32View(newBuffer);
+  HEAPU8 = new Uint8View(newBuffer);
+  HEAPU16 = new Uint16View(newBuffer);
+  HEAPU32 = new Uint32View(newBuffer);
+  HEAPF32 = new Float32View(newBuffer);
+  HEAPF64 = new Float64View(newBuffer);
+  buffer = newBuffer;
+  return true;
+}
+'''] + ['''
+// EMSCRIPTEN_START_FUNCS
+'''] + runtime_funcs + funcs_js + ['''
   %s
 
   return %s;
@@ -1195,6 +1243,296 @@ Runtime.registerFunctions(%(sigs)s, Module);
       outfile.close()
       shared.try_delete(outfile.name) # remove partial output
 
+def emscript_wasm_backend(infile, settings, outfile, outfile_name, libraries=[], compiler_engine=None,
+                          temp_files=None, DEBUG=None, DEBUG_CACHE=None):
+  # Overview:
+  #   * Run LLVM backend to emit .s
+  #   * Run Binaryen's s2wasm to generate WebAssembly.
+  #   * We may also run some Binaryen passes here.
+
+  temp_s = temp_files.get('.wb.s').name
+  backend_compiler = os.path.join(shared.LLVM_ROOT, 'llc')
+  backend_args = [backend_compiler, infile, '-march=wasm32', '-filetype=asm', '-o', temp_s]
+  if DEBUG:
+    logging.debug('emscript: llvm wasm backend: ' + ' '.join(backend_args))
+    t = time.time()
+  shared.check_call(backend_args)
+  if DEBUG:
+    logging.debug('  emscript: llvm wasm backend took %s seconds' % (time.time() - t))
+    t = time.time()
+    import shutil
+    shutil.copyfile(temp_s, os.path.join(shared.CANONICAL_TEMP_DIR, 'emcc-llvm-backend-output.s'))
+
+  assert shared.BINARYEN_ROOT, 'need BINARYEN_ROOT config set so we can use Binaryen s2wasm on the backend output'
+  wasm = outfile_name[:-3] + '.wast'
+  s2wasm_args = [os.path.join(shared.BINARYEN_ROOT, 's2wasm'), temp_s]
+  s2wasm_args += ['--global-base=%d' % shared.Settings.GLOBAL_BASE]
+  if DEBUG:
+    logging.debug('emscript: binaryen s2wasm: ' + ' '.join(s2wasm_args))
+    t = time.time()
+    s2wasm_args += ['--debug']
+  shared.check_call(s2wasm_args, stdout=open(wasm, 'w'))
+  if DEBUG:
+    logging.debug('  emscript: binaryen s2wasm took %s seconds' % (time.time() - t))
+    t = time.time()
+    import shutil
+    shutil.copyfile(wasm, os.path.join(shared.CANONICAL_TEMP_DIR, 'emcc-s2wasm-output.wast'))
+
+  # js compiler
+
+  if DEBUG: logging.debug('emscript: js compiler glue')
+
+  # Integrate info from backend
+
+  output = open(wasm).read()
+  parts = output.split('\n;; METADATA:')
+  assert len(parts) == 2
+  metadata_raw = parts[1]
+  parts = output = None
+
+  if DEBUG: logging.debug("METAraw %s", metadata_raw)
+  try:
+    metadata_json = json.loads(metadata_raw)
+  except Exception, e:
+    logging.error('emscript: failure to parse metadata output from s2wasm. raw output is: \n' + metadata_raw)
+    raise e
+
+  metadata = {
+    'declares': [],
+    'implementedFunctions': [],
+    'externs': [],
+    'simd': False,
+    'maxGlobalAlign': 0,
+    'initializers': [],
+    'exports': [],
+  }
+
+  for k, v in metadata_json.iteritems():
+    metadata[k] = v
+
+  # TODO: emit it from s2wasm; for now, we parse it right here
+  for line in open(wasm).readlines():
+    if line.startswith('  (import '):
+      parts = line.split(' ')
+      metadata['declares'].append(parts[3][1:])
+    elif line.startswith('  (func '):
+      parts = line.split(' ')
+      func = parts[3][1:]
+      metadata['implementedFunctions'].append(func)
+    elif line.startswith('  (export '):
+      parts = line.split(' ')
+      metadata['exports'].append('_' + parts[3][1:-1])
+
+  if DEBUG: logging.debug(repr(metadata))
+
+  settings['DEFAULT_LIBRARY_FUNCS_TO_INCLUDE'] = list(
+    set(settings['DEFAULT_LIBRARY_FUNCS_TO_INCLUDE'] + map(shared.JS.to_nice_ident, metadata['declares'])).difference(
+      map(lambda x: x[1:], metadata['implementedFunctions'])
+    )
+  ) + map(lambda x: x[1:], metadata['externs'])
+  if metadata['simd']:
+    settings['SIMD'] = 1
+
+  settings['MAX_GLOBAL_ALIGN'] = metadata['maxGlobalAlign']
+
+  # Save settings to a file to work around v8 issue 1579
+  settings_file = temp_files.get('.txt').name
+  def save_settings():
+    global settings_text
+    settings_text = json.dumps(settings, sort_keys=True)
+    s = open(settings_file, 'w')
+    s.write(settings_text)
+    s.close()
+  save_settings()
+
+  # Call js compiler
+  if DEBUG: t = time.time()
+  out = jsrun.run_js(path_from_root('src', 'compiler.js'), compiler_engine,
+                     [settings_file] + libraries, stdout=subprocess.PIPE, stderr=STDERR_FILE,
+                     cwd=path_from_root('src'), error_limit=300)
+  assert '//FORWARDED_DATA:' in out, 'Did not receive forwarded data in pre output - process failed?'
+  glue, forwarded_data = out.split('//FORWARDED_DATA:')
+
+  if DEBUG:
+    logging.debug('  emscript: glue took %s seconds' % (time.time() - t))
+    t = time.time()
+
+  last_forwarded_json = forwarded_json = json.loads(forwarded_data)
+
+  pre, post = glue.split('// EMSCRIPTEN_END_FUNCS')
+
+  # memory and global initializers
+
+  global_initializers = str(', '.join(map(lambda i: '{ func: function() { %s() } }' % i, metadata['initializers'])))
+
+  staticbump = metadata['staticBump']
+  while staticbump % 16 != 0: staticbump += 1
+  pre = pre.replace('STATICTOP = STATIC_BASE + 0;', '''STATICTOP = STATIC_BASE + %d;%s
+/* global initializers */ %s __ATINIT__.push(%s);
+''' % (staticbump,
+         'assert(STATICTOP < SPLIT_MEMORY, "SPLIT_MEMORY size must be big enough so the entire static memory, need " + STATICTOP);' if settings['SPLIT_MEMORY'] else '',
+         'if (!ENVIRONMENT_IS_PTHREAD)' if settings['USE_PTHREADS'] else '',
+         global_initializers))
+
+  pre = pre.replace('{{{ STATIC_BUMP }}}', str(staticbump))
+
+  # merge forwarded data
+  settings['EXPORTED_FUNCTIONS'] = forwarded_json['EXPORTED_FUNCTIONS']
+  all_exported_functions = set(shared.expand_response(settings['EXPORTED_FUNCTIONS'])) # both asm.js and otherwise
+
+  for additional_export in settings['DEFAULT_LIBRARY_FUNCS_TO_INCLUDE']: # additional functions to export from asm, if they are implemented
+    all_exported_functions.add('_' + additional_export)
+  exported_implemented_functions = set(metadata['exports'])
+  export_bindings = settings['EXPORT_BINDINGS']
+  export_all = settings['EXPORT_ALL']
+  all_implemented = metadata['implementedFunctions'] + forwarded_json['Functions']['implementedFunctions'].keys() # XXX perf?
+  for key in all_implemented:
+    if key in all_exported_functions or export_all or (export_bindings and key.startswith('_emscripten_bind')):
+      exported_implemented_functions.add(key)
+  implemented_functions = set(metadata['implementedFunctions'])
+  if settings['ASSERTIONS'] and settings.get('ORIGINAL_EXPORTED_FUNCTIONS'):
+    original_exports = settings['ORIGINAL_EXPORTED_FUNCTIONS']
+    if original_exports[0] == '@': original_exports = json.loads(open(original_exports[1:]).read())
+    for requested in original_exports:
+      # check if already implemented
+      # special-case malloc, EXPORTED by default for internal use, but we bake in a trivial allocator and warn at runtime if used in ASSERTIONS \
+      if requested not in all_implemented and \
+         requested != '_malloc' and \
+         (('function ' + requested.encode('utf-8')) not in pre): # could be a js library func
+        logging.warning('function requested to be exported, but not implemented: "%s"', requested)
+
+  asm_consts = [0]*len(metadata['asmConsts'])
+  all_sigs = []
+  for k, v in metadata['asmConsts'].iteritems():
+    const = v[0].encode('utf-8')
+    sigs = v[1]
+    if const[0] == '"' and const[-1] == '"':
+      const = const[1:-1]
+    const = '{ ' + const + ' }'
+    args = []
+    arity = max(map(len, sigs)) - 1
+    for i in range(arity):
+      args.append('$' + str(i))
+    const = 'function(' + ', '.join(args) + ') ' + const
+    asm_consts[int(k)] = const
+    all_sigs += sigs
+
+  asm_const_funcs = []
+  for sig in set(all_sigs):
+    forwarded_json['Functions']['libraryFunctions']['_emscripten_asm_const_' + sig] = 1
+    args = ['a%d' % i for i in range(len(sig)-1)]
+    all_args = ['code'] + args
+    asm_const_funcs.append(r'''
+function _emscripten_asm_const_%s(%s) {
+return ASM_CONSTS[code](%s);
+}''' % (sig.encode('utf-8'), ', '.join(all_args), ', '.join(args)))
+
+  pre = pre.replace('// === Body ===', '// === Body ===\n' + '\nvar ASM_CONSTS = [' + ',\n '.join(asm_consts) + '];\n' + '\n'.join(asm_const_funcs) + '\n')
+
+  outfile.write(pre)
+  pre = None
+
+  basic_funcs = ['abort', 'assert']
+
+  asm_runtime_funcs = ['stackAlloc', 'stackSave', 'stackRestore', 'establishStackSpace', 'setThrew']
+  asm_runtime_funcs += ['setTempRet0', 'getTempRet0']
+
+  if settings['ONLY_MY_CODE']:
+    asm_runtime_funcs = []
+
+  def quote(prop):
+    if settings['USE_CLOSURE_COMPILER'] == 2:
+      return "'" + prop + "'"
+    else:
+      return prop
+
+  def access_quote(prop):
+    if settings['USE_CLOSURE_COMPILER'] == 2:
+      return "['" + prop + "']"
+    else:
+      return '.' + prop
+
+  # calculate globals
+  try:
+    del forwarded_json['Variables']['globals']['_llvm_global_ctors'] # not a true variable
+  except:
+    pass
+  if not settings['RELOCATABLE']:
+    global_vars = metadata['externs']
+  else:
+    global_vars = [] # linkable code accesses globals through function calls
+  global_funcs = list(set([key for key, value in forwarded_json['Functions']['libraryFunctions'].iteritems() if value != 2]).difference(set(global_vars)).difference(implemented_functions))
+  def math_fix(g):
+    return g if not g.startswith('Math_') else g.split('_')[1]
+
+  basic_vars = ['STACKTOP', 'STACK_MAX', 'ABORT']
+  basic_float_vars = []
+
+  # sent data
+  the_global = '{}'
+  sending = '{ ' + ', '.join(['"' + math_fix(s) + '": ' + s for s in basic_funcs + global_funcs + basic_vars + basic_float_vars + global_vars]) + ' }'
+  # received
+  receiving = ''
+  if settings['ASSERTIONS']:
+    # assert on the runtime being in a valid state when calling into compiled code. The only exceptions are
+    # some support code
+    receiving = '\n'.join(['var real_' + s + ' = asm["' + s + '"]; asm["' + s + '''"] = function() {
+assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
+assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
+return real_''' + s + '''.apply(null, arguments);
+};
+''' for s in exported_implemented_functions if s not in ['_memcpy', '_memset', 'runPostSets', '_emscripten_replace_memory']])
+
+  if not settings['SWAPPABLE_ASM_MODULE']:
+    receiving += ';\n'.join(['var ' + s + ' = Module["' + s + '"] = asm["' + s[1:] + '"]' for s in exported_implemented_functions])
+  else:
+    receiving += 'Module["asm"] = asm;\n' + ';\n'.join(['var ' + s + ' = Module["' + s + '"] = function() { return Module["asm"]["' + s[1:] + '"].apply(null, arguments) }' for s in exported_implemented_functions])
+  receiving += ';\n'
+
+  # finalize
+
+  if settings['USE_PTHREADS']:
+    shared_array_buffer = "if (typeof SharedArrayBuffer !== 'undefined') Module.asmGlobalArg['Atomics'] = Atomics;"
+  else:
+    shared_array_buffer = ''
+
+  runtime_funcs = []
+
+  funcs_js = ['''
+Module%s = %s;
+%s
+Module%s = %s;
+''' % (access_quote('asmGlobalArg'), the_global,
+     shared_array_buffer,
+     access_quote('asmLibraryArg'), sending) + '''
+var asm = Module['asm'](%s, %s, buffer);
+%s;
+''' % ('Module' + access_quote('asmGlobalArg'),
+     'Module' + access_quote('asmLibraryArg'),
+     receiving)]
+
+  funcs_js.append('''
+Runtime.stackAlloc = asm['stackAlloc'];
+Runtime.stackSave = asm['stackSave'];
+Runtime.stackRestore = asm['stackRestore'];
+Runtime.establishStackSpace = asm['establishStackSpace'];
+''')
+
+  funcs_js.append('''
+Runtime.setTempRet0 = asm['setTempRet0'];
+Runtime.getTempRet0 = asm['getTempRet0'];
+''')
+
+  for i in range(len(funcs_js)): # do this loop carefully to save memory
+    if WINDOWS: funcs_js[i] = funcs_js[i].replace('\r\n', '\n') # Normalize to UNIX line endings, otherwise writing to text file will duplicate \r\n to \r\r\n!
+    outfile.write(funcs_js[i])
+  funcs_js = None
+
+  if WINDOWS: post = post.replace('\r\n', '\n') # Normalize to UNIX line endings, otherwise writing to text file will duplicate \r\n to \r\r\n!
+  outfile.write(post)
+
+  outfile.close()
+
 if os.environ.get('EMCC_FAST_COMPILER') == '0':
   logging.critical('Non-fastcomp compiler is no longer available, please use fastcomp or an older version of emscripten')
   sys.exit(1)
@@ -1217,8 +1555,10 @@ def main(args, compiler_engine, cache, temp_files, DEBUG, DEBUG_CACHE):
     shared.Building.ensure_struct_info(struct_info)
     if DEBUG: logging.debug('  emscript: bootstrapping struct info complete')
 
-  emscript(args.infile, settings, args.outfile, libraries, compiler_engine=compiler_engine,
-           temp_files=temp_files, DEBUG=DEBUG, DEBUG_CACHE=DEBUG_CACHE)
+  emscripter = emscript_wasm_backend if settings['WASM_BACKEND'] else emscript
+
+  emscripter(args.infile, settings, args.outfile, args.outfile_name, libraries, compiler_engine=compiler_engine,
+             temp_files=temp_files, DEBUG=DEBUG, DEBUG_CACHE=DEBUG_CACHE)
 
 def _main(args=None):
   if args is None:
@@ -1291,6 +1631,7 @@ WARNING: You should normally never use this! Use emcc instead.
   if len(positional) != 1:
     raise RuntimeError('Must provide exactly one positional argument. Got ' + str(len(positional)) + ': "' + '", "'.join(positional) + '"')
   keywords.infile = os.path.abspath(positional[0])
+  keywords.outfile_name = keywords.outfile
   if isinstance(keywords.outfile, basestring):
     keywords.outfile = open(keywords.outfile, 'w')
 
@@ -1325,4 +1666,3 @@ WARNING: You should normally never use this! Use emcc instead.
 
 if __name__ == '__main__':
   _main()
-
